/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.dataawarereplayer.precision;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.Set;

import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;

import org.processmining.framework.util.ui.widgets.ProMCheckComboBox;
import org.processmining.framework.util.ui.widgets.ProMComboBox;
import org.processmining.framework.util.ui.widgets.ProMPropertiesExpertModePanel;
import org.processmining.framework.util.ui.widgets.WidgetColors;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;

import com.fluxicon.slickerbox.components.NiceIntegerSlider;
import com.fluxicon.slickerbox.components.NiceSlider.Orientation;
import com.fluxicon.slickerbox.factory.SlickerFactory;
import com.google.common.base.Joiner;
import com.google.common.collect.ArrayTable;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Iterables;
import com.google.common.collect.SetMultimap;
import com.google.common.collect.Sets;

/**
 * @author F. Mannhardt
 * 
 */
class PrecisionConfigPanel extends ProMPropertiesExpertModePanel {

	private static final String TRANSITION_COLUMN = "Transition";
	private static final String RESOURCES_COLUMN = "Possible Resources";
	
	private static final long serialVersionUID = 7108370292413921996L;

	public enum PrecisionType {
		ACTIVITY, RESOURCE
	}

	private NiceIntegerSlider concurrentThreads;

	private ProMComboBox<PrecisionType> precisionMeasure;
	
	private JCheckBox debugMode;
	private JCheckBox penalizeImpreciseEndings;
	private JCheckBox assumeUnassignedVariablesFree;

	private ProMCheckComboBox relevantAttributes;

	private JTable resourceConfiguration;
	private ArrayTable<Transition, String, String> resourceTable;

	public PrecisionConfigPanel(Collection<Transition> transitions,
			SetMultimap<Transition, String> activityExecutorMap, Set<String> attributesInLog) {
		super("", true, "Expert Options");
		setPreferredSize(new Dimension(200, 300));
		setBackground(WidgetColors.PROPERTIES_BACKGROUND);
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		SlickerFactory sF = SlickerFactory.instance();
		
		precisionMeasure = addComboBox("Measure", PrecisionType.values(), false);

		int availableProcessors = Runtime.getRuntime().availableProcessors();
		concurrentThreads = sF.createNiceIntegerSlider("", 1, availableProcessors, Math.max(1, availableProcessors / 2),
				Orientation.HORIZONTAL);
		addProperty("# Concurrent Threads", concurrentThreads);

		relevantAttributes = new ProMCheckComboBox(attributesInLog);
		addProperty("Relevant Attributes", relevantAttributes, false);
		addProperty(
				"",
				sF.createLabel("Select which additional attributes from the log are relevant for the precision measurement"),
				false);
		
		resourceTable = ArrayTable.create(transitions, ImmutableList.of(TRANSITION_COLUMN, RESOURCES_COLUMN));

		TableModel resourceModel = new AbstractTableModel() {

			private static final long serialVersionUID = 5022437081512803819L;

			public boolean isCellEditable(int row, int column) {
				return column != 0;
			}

			public int getRowCount() {
				return resourceTable.rowKeySet().size();
			}

			public int getColumnCount() {
				return resourceTable.columnKeySet().size();
			}

			public Object getValueAt(int rowIndex, int columnIndex) {
				return resourceTable.at(rowIndex, columnIndex);
			}

			public String getColumnName(int column) {
				return resourceTable.columnKeyList().get(column);
			}

			public Class<?> getColumnClass(int columnIndex) {
				return String.class;
			}

			public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
				resourceTable.set(rowIndex, columnIndex, (String) aValue);
			}

		};
		
		for (Transition t : transitions) {
			resourceTable.put( t, TRANSITION_COLUMN, t.getLabel() );
			resourceTable.put( t, RESOURCES_COLUMN, createDefaultExecutorSet(activityExecutorMap.get(t)) );
		}

		resourceConfiguration = new JTable(resourceModel);
		JScrollPane scrollWrapper = new JScrollPane(resourceConfiguration);
		scrollWrapper.setPreferredSize(new Dimension(Short.MAX_VALUE, 150));
		addProperty("Resources", scrollWrapper);
		addProperty("", sF.createLabel("Specify which resources (attribute org:resource) may execute activities (comma separated list)"), true);

		debugMode = addCheckBox("Debug Mode", false, true);
		addProperty("", sF.createLabel("Enable debug mode with assertions resulting in much slower computation"), true);

		penalizeImpreciseEndings = addCheckBox("Detect Imprecise Ending", false, true);
		addProperty("", sF.createLabel("Detect imprecisions in final states at the end of traces"), true);
		
		assumeUnassignedVariablesFree = addCheckBox("Treat Unassigned Variables as Free", true, true);
		addProperty("", sF.createLabel("Assumes that guards refering to unassigned variables can be satisfied"), true);

		precisionMeasure.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				resourceConfiguration.getParent().getParent().getParent()
						.setVisible(precisionMeasure.getSelectedItem() == PrecisionType.RESOURCE);
			}
		});
		precisionMeasure.setSelectedItem(PrecisionType.ACTIVITY);
	}

	private String createDefaultExecutorSet(Set<String> resources) {
		return Joiner.on(',').join(resources);
	}

	public PrecisionType getPrecisionMeasure() {
		return (PrecisionType) precisionMeasure.getSelectedItem();
	}

	public boolean getIsDebugMode() {
		return debugMode.getModel().isSelected();
	}

	public boolean getIsPenalizeImpreciseEndings() {
		return penalizeImpreciseEndings.getModel().isSelected();
	}
	
	public boolean getIsAssumeUnassignedVariablesFree() {
		return assumeUnassignedVariablesFree.getModel().isSelected();
	}

	public int getConcurrentThreads() {
		return concurrentThreads.getValue();
	}

	public Set<String> getRelevantAttributes() {
		if (relevantAttributes.getSelectedItems() != null) {
			return Sets.newHashSet(Iterables.filter(relevantAttributes.getSelectedItems(), String.class));	
		} else {
			return Collections.emptySet();
		}
	}

	public Map<Transition, String> getExecutors() {
		return resourceTable.column(resourceTable.columnKeyList().get(1));
	}

}
