package org.processmining.dataawarereplayer.precision.measure;

import java.text.MessageFormat;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.extension.std.XOrganizationalExtension;
import org.processmining.dataawarereplayer.precision.PrecisionConfig;
import org.processmining.dataawarereplayer.precision.PrecisionMeasureException;
import org.processmining.dataawarereplayer.precision.projection.ProcessProjector;
import org.processmining.dataawarereplayer.precision.projection.ProjectedEvent;
import org.processmining.dataawarereplayer.precision.projection.ProjectedTrace;
import org.processmining.dataawarereplayer.precision.state.DPNStateImpl;
import org.processmining.dataawarereplayer.precision.state.MarkingAwareModelState;
import org.processmining.dataawarereplayer.precision.state.PrecisionState;
import org.processmining.dataawarereplayer.precision.state.StateFactory;
import org.processmining.dataawarereplayer.precision.state.StateTransition;
import org.processmining.dataawarereplayer.precision.state.TraceStatePrefixImpl;
import org.processmining.dataawarereplayer.precision.state.TransitionAwareStateTransition;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.semantics.petrinet.EfficientPetrinetSemantics;
import org.processmining.models.semantics.petrinet.EfficientPetrinetSemantics.PlaceVisitor;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.impl.EfficientPetrinetSemanticsImpl;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataState;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateFactoryImpl;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateHelper;
import org.processmining.plugins.balancedconformance.dataflow.util.ExpressionEvaluator;

import com.google.common.base.Predicate;
import com.google.common.collect.ConcurrentHashMultiset;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSetMultimap;
import com.google.common.collect.Multimaps;
import com.google.common.collect.Multiset;
import com.google.common.collect.Multiset.Entry;
import com.google.common.collect.SetMultimap;

public abstract class AbstractDPNPrecisionMeasure implements PrecisionMeasure {

	private static final class PrecisionInfo {
		private final SetMultimap<PrecisionState, StateTransition> observedBehavior = Multimaps
				.<PrecisionState, StateTransition>synchronizedSetMultimap(
						HashMultimap.<PrecisionState, StateTransition>create());
		private final SetMultimap<PrecisionState, StateTransition> possibleBehavior = Multimaps
				.<PrecisionState, StateTransition>synchronizedSetMultimap(
						HashMultimap.<PrecisionState, StateTransition>create());
		private final Multiset<PrecisionState> stateOccurence = ConcurrentHashMultiset.create();

		public void addStateOccurence(PrecisionState currentState) {
			stateOccurence.add(currentState);
		}

		public void addObservedBehavior(PrecisionState currentState, StateTransition newStateTransition) {
			observedBehavior.put(currentState, newStateTransition);
		}

		public void addPossibleBehavior(PrecisionState currentState, Collection<StateTransition> possibleInState) {
			possibleBehavior.putAll(currentState, possibleInState);
		}

		public void addPossibleBehavior(PrecisionState currentState, StateTransition transition) {
			possibleBehavior.put(currentState, transition);
		}

		public Multiset<PrecisionState> getObservedStates() {
			return observedBehavior.keys();
		}

		public Multiset<PrecisionState> getPossibleStates() {
			return possibleBehavior.keys();
		}

		public Set<Entry<PrecisionState>> getStates() {
			return stateOccurence.entrySet();
		}

		public SetMultimap<PrecisionState, StateTransition> getObservedBehavior() {
			return observedBehavior;
		}

		public SetMultimap<PrecisionState, StateTransition> getPossibleBehavior() {
			return possibleBehavior;
		}

		public int getFrequency(PrecisionState state) {
			return stateOccurence.count(state);
		}

	}

	private static final String NO_MORE_TRANSITIONS_ENABLED = "Trace cannot be replayed on the model, this is an error on the used process projection: %s!";

	private static final String VARIABLE_ASSIGNMENT_WRONG = "Trace %s cannot be replayed on the model, transition %s is not enabled because the guard %s is not fulfilled by variable assignment %s. This is an error on the used process projection: %s!";

	private static final String TRANSITION_NOT_ENABLED_ERROR = "Trace cannot be replayed on the model, transition %s is not enabled. This is an error on the used process projection: %s!";

	public static class PrecisionResultPlace {

		double precision;
		long observed;
		long possible;

		public PrecisionResultPlace(double precision, long observed, long possible) {
			this.precision = precision;
			this.observed = observed;
			this.possible = possible;
		}

	}

	protected static final StateTransition ARTIFICIAL_END = new StateTransition() {

		public String toString() {
			return "ARTIFICIAL END";
		}

	};

	private final DataPetriNet net;
	protected final PrecisionConfig config;

	protected final ImmutableList<Transition> transitions;
	protected final Set<PNWDTransition> expressionContainsResourceAssignmentMap;
	private final ImmutableList<Place> places;
	private final EfficientPetrinetSemanticsImpl semanticsTemplate;

	private final DataStateFactory dataStateFactory;
	private final StateFactory stateFactory;

	protected final PrecisionInfo globalPrecisionInfo = new PrecisionInfo();
	private final Map<Place, PrecisionResultPlace> precisionAtPlaces = new HashMap<>();

	private final SetMultimap<Place, Transition> directSuccessors = HashMultimap.create();
	protected final Map<Place, PrecisionInfo> localPrecisionInfo = new HashMap<>();
	private final Map<Place, PrecisionResultPlace> localPrecisionAtPlaces = new HashMap<>();

	protected final ProcessProjector processProjector;

	public AbstractDPNPrecisionMeasure(PrecisionConfig config, ProcessProjector processProjector,
			StateFactory stateFactory, DataPetriNet net) {
		this.config = config;
		this.processProjector = processProjector;
		this.stateFactory = stateFactory;
		this.net = net;
		this.transitions = ImmutableList.copyOf(net.getTransitions());
		this.places = ImmutableList.copyOf(net.getPlaces());
		this.semanticsTemplate = new EfficientPetrinetSemanticsImpl(net, config.getInitialMarking());
		this.dataStateFactory = new DataStateFactoryImpl(processProjector.getAttributeType());

		this.expressionContainsResourceAssignmentMap = new HashSet<>();
		for (Transition t : net.getTransitions()) {
			if (t instanceof PNWDTransition) {
				PNWDTransition transition = (PNWDTransition) t;
				if (transition.getGuardExpression() != null) {
					GuardExpression guard = transition.getGuardExpression();
					for (String primeVariables : guard.getPrimeVariables()) {
						if (primeVariables.equals(XOrganizationalExtension.KEY_RESOURCE)) {
							expressionContainsResourceAssignmentMap.add(transition);
						}
					}
				}
			}
		}

		for (Place p : net.getPlaces()) {
			localPrecisionInfo.put(p, new PrecisionInfo());
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : net.getOutEdges(p)) {
				if (edge.getTarget() instanceof Transition) {
					directSuccessors.put(p, (Transition) edge.getTarget());
				}
			}
		}
	}

	public Object getModel() {
		return net;
	}

	protected abstract StateTransition newStateTransition(ProjectedEvent event);

	protected abstract Collection<StateTransition> recordPossibilities(DataState variableAssignment,
			PrecisionState state, EfficientPetrinetSemantics petriNetSemantics) throws PrecisionMeasureException;

	public void processTrace(ProjectedTrace projectedTrace) throws PrecisionMeasureException {

		EfficientPetrinetSemantics petriNetSemantics = new EfficientPetrinetSemanticsImpl(semanticsTemplate);
		DataState variableAssignment = dataStateFactory.newDataState();

		PrecisionState currentState = stateFactory.newPrecisionState(
				new TraceStatePrefixImpl(Collections.<ProjectedEvent>emptyList()),
				new DPNStateImpl(petriNetSemantics.getState(), variableAssignment.clone(false).makeStateImmutable()));

		for (ListIterator<ProjectedEvent> iterator = projectedTrace.listIterator(); iterator.hasNext();) {
			final ProjectedEvent event = iterator.next();
			final Transition currentTransition = event.getTransition();

			// Global observations
			globalPrecisionInfo.addStateOccurence(currentState);
			StateTransition newStateTransition = newStateTransition(event);
			globalPrecisionInfo.addObservedBehavior(currentState, newStateTransition);
			// Local observations
			processLocalObservations(petriNetSemantics, currentState, currentTransition, newStateTransition);

			Collection<StateTransition> possibleInState = recordPossibilities(variableAssignment, currentState,
					petriNetSemantics);
			if (possibleInState.isEmpty()) {
				throw new PrecisionMeasureException(String.format(NO_MORE_TRANSITIONS_ENABLED, projectedTrace));
			}

			// Global possibilities
			globalPrecisionInfo.addPossibleBehavior(currentState, possibleInState);
			// Local possibilities
			processLocalPossibilites(petriNetSemantics, currentState, currentTransition, possibleInState);

			for (String attributeName : event.getAttributesWritten()) {
				Object varValue = event.getAttributeValue(attributeName);
				DataStateHelper.storeValue(variableAssignment, attributeName.concat("'"), varValue,
						processProjector.getAttributeType().get(attributeName),
						processProjector.getStringDiscretizer());
			}

			if (config.isDebug()) {
				if (petriNetSemantics.isEnabled(currentTransition)) {
					petriNetSemantics.directExecuteExecutableTransition(currentTransition);
					// ASSERT: Guard is fulfilled!
					if (currentTransition instanceof PNWDTransition
							&& ((PNWDTransition) currentTransition).getGuardExpression() != null) {
						GuardExpression guardExpression = ((PNWDTransition) currentTransition).getGuardExpression();
						if (!ExpressionEvaluator.failSafeIsTrue(guardExpression, variableAssignment,
								processProjector.getAttributeType(), processProjector.getStringDiscretizer())) {
							throw new PrecisionMeasureException(
									String.format(VARIABLE_ASSIGNMENT_WRONG, projectedTrace.getName(),
											currentTransition, guardExpression, variableAssignment, projectedTrace));
						}
					}
				} else {
					throw new PrecisionMeasureException(
							String.format(TRANSITION_NOT_ENABLED_ERROR, currentTransition, projectedTrace));
				}
			} else {
				// Assuming that the process projection is OK
				petriNetSemantics.directExecuteExecutableTransition(currentTransition);
			}

			variableAssignment = variableAssignment.clone(true);

			currentState = stateFactory.newPrecisionState(
					new TraceStatePrefixImpl(projectedTrace.subList(0, iterator.nextIndex())), new DPNStateImpl(
							petriNetSemantics.getState(), variableAssignment.clone(false).makeStateImmutable()));

			if (config.isPenalizeImpreciseEnding()) {
				if (!iterator.hasNext()) {
					// Check precision in the 'final' state
					globalPrecisionInfo.addStateOccurence(currentState);
					globalPrecisionInfo.addObservedBehavior(currentState, ARTIFICIAL_END);
					possibleInState = recordPossibilities(variableAssignment, currentState, petriNetSemantics);
					possibleInState.add(ARTIFICIAL_END);
					globalPrecisionInfo.addPossibleBehavior(currentState, possibleInState);

					processLocalObservations(petriNetSemantics, currentState, currentTransition, ARTIFICIAL_END);
					processLocalPossibilites(petriNetSemantics, currentState, currentTransition, possibleInState);
				}
			}
		}

	}

	private final void processLocalPossibilites(EfficientPetrinetSemantics petriNetSemantics,
			final PrecisionState currentState, final Transition currentTransition,
			final Collection<StateTransition> possibleInState) {
		petriNetSemantics.forEachMarkedPlace(new PlaceVisitor() {

			public void accept(Place place, int numTokens) {
				Set<Transition> sucessors = directSuccessors.get(place);
				if (sucessors.contains(currentTransition)) {
					// We remove a token directly from this place
					for (StateTransition transition : possibleInState) {
						if (transition instanceof TransitionAwareStateTransition) {
							if (sucessors.contains(((TransitionAwareStateTransition) transition).getActivity())) {
								localPrecisionInfo.get(place).addPossibleBehavior(currentState, transition);
							}
						}
					}
				}
			}
		});
	}

	private final void processLocalObservations(EfficientPetrinetSemantics petriNetSemantics,
			final PrecisionState currentState, final Transition currentTransition,
			final StateTransition newStateTransition) {
		petriNetSemantics.forEachMarkedPlace(new PlaceVisitor() {

			public void accept(Place place, int numTokens) {
				if (directSuccessors.get(place).contains(currentTransition)) {
					// We remove a token directly from this place
					PrecisionInfo localInfo = localPrecisionInfo.get(place);
					localInfo.addStateOccurence(currentState);
					localInfo.addObservedBehavior(currentState, newStateTransition);
				}
			}
		});
	}

	@Override
	public void collectResults(ResultMode mode) {
		if (mode == ResultMode.PLACE_LEVEL) {
			for (Place place : places) {
				calculatePlacePrecision(place);
			}
		}
	}

	@Override
	public double getPrecision() {
		long observed = 0;
		long possible = 0;

		Multiset<PrecisionState> observedMultiset = globalPrecisionInfo.getObservedStates();
		Multiset<PrecisionState> possibleMultiset = globalPrecisionInfo.getPossibleStates();

		for (Entry<PrecisionState> stateEntry : globalPrecisionInfo.getStates()) {

			final int occurence = stateEntry.getCount();
			final PrecisionState element = stateEntry.getElement();

			int observedStateTransitions = observedMultiset.count(element);
			int possibleStateTransitions = possibleMultiset.count(element);

			observed += observedStateTransitions * occurence;
			possible += possibleStateTransitions * occurence;
		}

		return (double) observed / (double) possible;
	}

	@Override
	public double getPrecision(Place p) {
		PrecisionResultPlace placePrecision = precisionAtPlaces.get(p);
		if (placePrecision == null) {
			return calculatePlacePrecision(p).precision;
		} else {
			return placePrecision.precision;
		}
	}

	@Override
	public long getObservedBehavior(Place p) {
		PrecisionResultPlace placePrecision = precisionAtPlaces.get(p);
		if (placePrecision == null) {
			return calculatePlacePrecision(p).observed;
		} else {
			return placePrecision.observed;
		}
	}

	@Override
	public long getPossibleBehavior(Place p) {
		PrecisionResultPlace placePrecision = precisionAtPlaces.get(p);
		if (placePrecision == null) {
			return calculatePlacePrecision(p).possible;
		} else {
			return placePrecision.possible;
		}
	}

	@Override
	public double getLocalPrecision(Place p) {
		PrecisionResultPlace placePrecision = localPrecisionAtPlaces.get(p);
		if (placePrecision == null) {
			return calculateLocalPlacePrecision(p).precision;
		} else {
			return placePrecision.precision;
		}
	}

	@Override
	public long getObservedLocalBehavior(Place p) {
		PrecisionResultPlace placePrecision = localPrecisionAtPlaces.get(p);
		if (placePrecision == null) {
			return calculateLocalPlacePrecision(p).observed;
		} else {
			return placePrecision.observed;
		}
	}

	@Override
	public long getPossibleLocalBehavior(Place p) {
		PrecisionResultPlace placePrecision = localPrecisionAtPlaces.get(p);
		if (placePrecision == null) {
			return calculateLocalPlacePrecision(p).possible;
		} else {
			return placePrecision.possible;
		}
	}

	@Override
	public SetMultimap<PrecisionState, StateTransition> getObservedStateTransitions(final Place p) {
		return ImmutableSetMultimap.copyOf(
				Multimaps.filterKeys(globalPrecisionInfo.getObservedBehavior(), new Predicate<PrecisionState>() {

					public boolean apply(PrecisionState state) {
						if (state.getModelState() instanceof MarkingAwareModelState) {
							if (semanticsTemplate
									.isMarked(((MarkingAwareModelState) state.getModelState()).getMarking(), p)) {
								return true;
							}
						}
						return false;
					}

				}));
	}

	@Override
	public SetMultimap<PrecisionState, StateTransition> getObservedStateTransitions() {
		return globalPrecisionInfo.getObservedBehavior();
	}

	@Override
	public SetMultimap<PrecisionState, StateTransition> getPossibleStateTransitions(final Place p) {
		return ImmutableSetMultimap.copyOf(
				Multimaps.filterKeys(globalPrecisionInfo.getPossibleBehavior(), new Predicate<PrecisionState>() {

					public boolean apply(PrecisionState state) {
						if (state.getModelState() instanceof MarkingAwareModelState) {
							if (semanticsTemplate
									.isMarked(((MarkingAwareModelState) state.getModelState()).getMarking(), p)) {
								return true;
							}
						}
						return false;
					}

				}));
	}

	@Override
	public SetMultimap<PrecisionState, StateTransition> getPossibleStateTransitions() {
		return globalPrecisionInfo.getPossibleBehavior();
	}

	@Override
	public long getFrequency(PrecisionState state) {
		return globalPrecisionInfo.getFrequency(state);
	}

	private PrecisionResultPlace calculateLocalPlacePrecision(Place place) {
		long observed = calculateObservedBehavior(localPrecisionInfo.get(place), place);
		long possible = calculatePossibleBehavior(localPrecisionInfo.get(place), place);

		if (possible > 0) {
			double precision = (double) observed / (double) possible;
			PrecisionResultPlace result = new PrecisionResultPlace(precision, observed, possible);
			localPrecisionAtPlaces.put(place, result);
			return result;
		} else {
			PrecisionResultPlace result = new PrecisionResultPlace(1.0, observed, possible);
			localPrecisionAtPlaces.put(place, result);
			return result;
		}
	}

	private PrecisionResultPlace calculatePlacePrecision(Place place) {
		long observed = calculateObservedBehavior(globalPrecisionInfo, place);
		long possible = calculatePossibleBehavior(globalPrecisionInfo, place);

		if (possible > 0) {
			double precision = (double) observed / (double) possible;
			PrecisionResultPlace result = new PrecisionResultPlace(precision, observed, possible);
			precisionAtPlaces.put(place, result);
			return result;
		} else {
			PrecisionResultPlace result = new PrecisionResultPlace(1.0, observed, possible);
			precisionAtPlaces.put(place, result);
			return result;
		}
	}

	private long calculateObservedBehavior(PrecisionInfo info, Place p) {
		long observed = 0;
		for (Entry<PrecisionState> entry : info.getObservedBehavior().keys().entrySet()) {
			observed += countBehaviorWithMarkedPlace(p, entry);
		}
		return observed;
	}

	private long calculatePossibleBehavior(PrecisionInfo info, Place p) {
		long possible = 0;
		for (Entry<PrecisionState> entry : info.getPossibleBehavior().keys().entrySet()) {
			possible += countBehaviorWithMarkedPlace(p, entry);
		}
		return possible;
	}

	private final int countBehaviorWithMarkedPlace(Place p, Entry<PrecisionState> stateEntry) {
		PrecisionState state = stateEntry.getElement();
		if (state.getModelState() instanceof MarkingAwareModelState) {
			if (semanticsTemplate.isMarked(((MarkingAwareModelState) state.getModelState()).getMarking(), p)) {
				return stateEntry.getCount() * globalPrecisionInfo.getFrequency(state);
			}
		}
		return 0;
	}

	@Override
	public double getPrecision(Marking marking) {
		long possibleContinuations = getPossibleContinuations(marking);
		if (possibleContinuations > 0) {
			return (double) getObservedContinuations(marking) / (double) possibleContinuations;
		} else {
			//Actually the measure is undefined, but we assume 1.0 for simplicity
			return 1.0;
		}
	}

	@Override
	public long getPossibleContinuations(Marking marking) {
		long possible = 0;
		for (Entry<PrecisionState> entry : globalPrecisionInfo.getPossibleBehavior().keys().entrySet()) {
			possible += countBehaviorWithLessEqualMarking(marking, entry);
		}
		return possible;
	}

	@Override
	public long getObservedContinuations(Marking marking) {
		long observed = 0;
		for (Entry<PrecisionState> entry : globalPrecisionInfo.getObservedBehavior().keys().entrySet()) {
			observed += countBehaviorWithLessEqualMarking(marking, entry);
		}
		return observed;
	}

	private final int countBehaviorWithLessEqualMarking(Marking marking, Entry<PrecisionState> stateEntry) {
		PrecisionState state = stateEntry.getElement();
		if (state.getModelState() instanceof MarkingAwareModelState) {
			if (semanticsTemplate.equalOrLessMarking(((MarkingAwareModelState) state.getModelState()).getMarking(),
					marking)) {
				return stateEntry.getCount() * globalPrecisionInfo.getFrequency(state);
			}
		}
		return 0;
	}

	@Override
	public String toString() {
		StringBuilder sBuilder = new StringBuilder();

		appendSummary(sBuilder);

		return sBuilder.toString();
	}

	public String debugString() {
		StringBuilder sBuilder = new StringBuilder();

		appendSummary(sBuilder);
		appendExtendedInformation(sBuilder);

		return sBuilder.toString();
	}

	private void appendSummary(StringBuilder sBuilder) {
		double precision = getPrecision();
		sBuilder.append(MessageFormat.format("Precision: {0}/{1}={2,number,#.###}\n",
				countBehavior(globalPrecisionInfo.getObservedBehavior().keys()),
				countBehavior(globalPrecisionInfo.getPossibleBehavior().keys()), precision));
		sBuilder.append("\n");
		sBuilder.append("Precision per Place:\n");
		for (Place place : places) {
			sBuilder.append(MessageFormat.format("{0}:{1,number,#.###}\n", place.getLabel(), getPrecision(place)));
		}
	}

	private void appendExtendedInformation(StringBuilder sBuilder) {
		sBuilder.append("\nObserved / Possible Behavior\n");

		Multiset<PrecisionState> observedMultiset = globalPrecisionInfo.getObservedBehavior().keys();
		Multiset<PrecisionState> possibleMultiset = globalPrecisionInfo.getPossibleBehavior().keys();

		for (Entry<PrecisionState> stateEntry : globalPrecisionInfo.getStates()) {
			int occurence = stateEntry.getCount();

			int observedStateTransitions = observedMultiset.count(stateEntry.getElement());
			int possibleStateTransitions = possibleMultiset.count(stateEntry.getElement());

			sBuilder.append(String.format("%s x (%s / %s) @ %s", occurence, observedStateTransitions,
					possibleStateTransitions, stateEntry.getElement().toString()));
			sBuilder.append("\n");
		}
	}

	private final int countBehavior(Multiset<PrecisionState> behaviorAtState) {
		int count = 0;
		for (Entry<PrecisionState> stateEntry : globalPrecisionInfo.getStates()) {
			count += behaviorAtState.count(stateEntry.getElement()) * stateEntry.getCount();
		}
		return count;
	}

}