package org.processmining.dataawarereplayer.precision.projection;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.log.utils.XUtils;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.plugins.DataConformance.Utility;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizer;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizerException;

import com.google.common.base.Function;
import com.google.common.collect.ForwardingList;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Iterators;
import com.google.common.collect.Lists;
import com.google.common.collect.SetMultimap;
import com.google.common.collect.Sets;
import com.google.common.collect.Sets.SetView;

public final class ProcessProjectorFittingImpl implements ProcessProjector {
	
	private static class ProjectedEventImpl implements ProjectedEvent {
		
		private final XEvent e;
		private final Transition transition;
		private final LogMapping logMapping;

		ProjectedEventImpl(XEvent e, Transition transition, LogMapping logMapping) {
			this.e = e;
			this.logMapping = logMapping;
			this.transition = transition;
		}

		public Object getAttributeValue(String varName) {
			return Utility.getValue(e.getAttributes().get(varName));
		}
		
		public Set<String> getAttributesWritten() {
			Set<String> variablesWritten = logMapping.getVariablesToWrite().get(transition);
			SetView<String> attributesInLog = Sets.intersection(logMapping.getExtraAttributes(), e.getAttributes().keySet());
			return Sets.union(variablesWritten, attributesInLog);
		}

		public Transition getTransition() {							
			return transition;
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((getTransition() == null) ? 0 : getTransition().hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (!(obj instanceof ProjectedEventImpl))
				return false;
			ProjectedEventImpl other = (ProjectedEventImpl) obj;
			if (getTransition() == null) {
				if (other.getTransition() != null)
					return false;
			} else if (!getTransition().equals(other.getTransition()))
				return false;
			return true;
		}

		public String toString() {
			return getTransition().getLabel();
		}
		
	}

	private static final class ProjectedTraceFittingImpl extends ForwardingList<ProjectedEvent> implements ProjectedTrace {

		private final List<ProjectedEvent> projecteEvents;
		private final XTrace logTrace;

		public ProjectedTraceFittingImpl(XTrace logTrace, final LogMapping logMapping) {
			this.logTrace = logTrace;
			projecteEvents = ImmutableList.copyOf(Lists.transform(logTrace, new Function<XEvent, ProjectedEvent>() {

				public ProjectedEvent apply(final XEvent e) {
					Set<Transition> transitionSet = logMapping.getEventIdentifyToTransition().get(logMapping.getEventClasses().getClassOf(e).getId());
					return new ProjectedEventImpl(e, transitionSet.iterator().next(), logMapping);
				}
			}));
		}

		protected List<ProjectedEvent> delegate() {
			return projecteEvents;
		}

		public String getName() {
			return XUtils.getConceptName(logTrace);
		}
		
	}

	private final XLog log;
	private final LogMapping logMapping;	

	@SuppressWarnings("rawtypes")
	public ProcessProjectorFittingImpl(DataPetriNet net, XLog log, XEventClasses eventClasses, Map<String, Class> extraAttributes, SetMultimap<Transition, String> executors) throws ProcessProjectionException {
		this.log = log;
		
		SetMultimap<String, Transition> eventIdentityToTransition = HashMultimap.create();
		for (Transition transition : net.getTransitions()) {
			for (XEventClass eventClass : eventClasses.getClasses()) {
				if (eventClass.getId().equals(transition.getLabel())) {
					eventIdentityToTransition.put(eventClass.getId(), transition);
				}
			}
		}
		
		try {
			Map<String, String> variableMapping = new HashMap<>();
			for (DataElement element: net.getVariables()) {
				variableMapping.put(element.getVarName(), element.getVarName());
			}
		
			logMapping = new LogMapping(net, log, eventClasses, eventIdentityToTransition, variableMapping, extraAttributes, executors);
			
		} catch (StringDiscretizerException e) {
			throw new ProcessProjectionException(e);
		}
	}

	public Iterator<ProjectedTrace> projectedTraces() {
		return Iterators.transform(log.iterator(), new Function<XTrace, ProjectedTrace>() {

			public ProjectedTrace apply(XTrace logTrace) {
				return new ProjectedTraceFittingImpl(logTrace, logMapping);
			}
		});
	}

	@SuppressWarnings("rawtypes")
	public Map<String, Class> getAttributeType() {
		return logMapping.getAttributeType();
	}

	public StringDiscretizer getStringDiscretizer() {
		return logMapping.getDiscretizer();
	}

}