/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.dataflow;

import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.balancedconformance.cache.DataViewBuilder;
import org.processmining.plugins.balancedconformance.cache.DataViewResultCache;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;
import org.processmining.plugins.balancedconformance.controlflow.override.DataAwareThread.PartialResultFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataState;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;
import org.processmining.plugins.balancedconformance.result.DataAlignedTrace;

import nl.tue.astar.Tail;

/**
 * {@link DataAlignmentAdapter} that checks if we already stored the result in a
 * cache.
 * 
 * @author F. Mannhardt
 * 
 */
public class DataAlignmentAdapterCacheImpl implements DataAlignmentAdapter, PartialDataAlignmentObserver {

	private static final int MAX_CACHE_SIZE = 256 * 1024;
	
	private final AbstractPDelegate<? extends Tail> delegate;
	private final LogMapping logMapping;

	private final DataStateFactory dataStateFactory;
	private final DataViewResultCache alignmentCache;
	private int cacheHits = 0;
	private boolean isCacheDirty = true;
	private boolean isCacheHit = false;

	public DataAlignmentAdapterCacheImpl(DataViewBuilder dataViewBuilder, DataStateFactory dataStateFactory,
			AbstractPDelegate<? extends Tail> delegate, LogMapping logMapping) {
		super();
		this.dataStateFactory = dataStateFactory;
		this.delegate = delegate;
		this.logMapping = logMapping;
		this.alignmentCache = new DataViewResultCache(dataViewBuilder, delegate, true, MAX_CACHE_SIZE);
	}

	public DataAlignedTrace retrieveFullDataAlignment(TraceMapping traceMapping,
			ControlFlowAlignmentResult controlFlowResult) throws DataAlignmentException {
		throw new UnsupportedOperationException(
				"This DataAlignment adapter does not support a complete data alignment!");
	}

	public PartialDataAlignedTrace retrievePartialDataAlignment(TraceMapping traceMapping,
			PartialResultFactory partialResultBuilder, DRecord currentRecord) {

		//TODO change to be more versatile and not requiring to be first in chain!
		alignmentCache.setCurrentResult(currentRecord);
		isCacheDirty = false;
		PartialDataAlignedTrace cachedAlignment = alignmentCache.getResult();
		if (cachedAlignment != null) {
			cacheHits++;
			isCacheHit = true;

			if (!hasWriteOps(currentRecord) && cachedAlignment != PartialDataAlignedTrace.IMPOSSIBLE) {
				// Clone the variable assignment, to move prime values as we are not writing variables
				DataState cachedDataState = cachedAlignment.getVariableAssignment();
				if (cachedDataState == null) {
					cachedDataState = traceMapping.getDataStateStore().get(cachedAlignment.getDataStateIndex());
				}
				if (cachedDataState != DataStateStore.EMPTY_IMMUTABLE_STATE) {
					return new PartialDataAlignedTrace(cachedAlignment.getDataCost(), cachedDataState.clone(true));
				}
			}

			return cachedAlignment;
		} else {
			return PartialDataAlignedTrace.NO_RESULT;
		}

	}

	public int getCacheHits() {
		return cacheHits;
	}

	public int getCacheSize() {
		return alignmentCache.size();
	}

	private boolean hasWriteOps(DRecord rec) {
		return logMapping.hasWriteOps(delegate.getTransition((short) rec.getModelMove()));
	}

	public void newPartialDataAlignment(DRecord record, PartialDataAlignedTrace dataState) {
		if (!isCacheHit && !isCacheDirty) {
			alignmentCache.putResult(dataState);
		}
		isCacheHit = false;
		isCacheDirty = true;
	}

	public DataStateStore newDataStateStore(int initialCapacity) {
		throw new UnsupportedOperationException("Data Cache does not support newDataStateStore");
	}

	public DataStateFactory getDataStateFactory() {
		return dataStateFactory;
	}

	public void close() {
		//NOop
	}

}