/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.dataflow;

import java.util.Map;
import java.util.Set;

import nl.tue.astar.AStarThread;
import nl.tue.astar.Tail;

import org.deckfour.xes.model.XEvent;
import org.processmining.datapetrinets.expression.Evaluator;
import org.processmining.datapetrinets.expression.FunctionProvider;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.expression.VariableProvider;
import org.processmining.datapetrinets.expression.syntax.ExprEqual;
import org.processmining.datapetrinets.expression.syntax.ExprRoot;
import org.processmining.datapetrinets.expression.syntax.ExprVariable;
import org.processmining.datapetrinets.expression.syntax.ExpressionVisitorException;
import org.processmining.datapetrinets.expression.syntax.Node;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;
import org.processmining.plugins.balancedconformance.controlflow.override.DataAwareThread.PartialResultFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataState;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateHelper;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.util.DataStateVariableProvider;
import org.processmining.plugins.balancedconformance.dataflow.util.ExpressionEvaluator;
import org.processmining.plugins.balancedconformance.functions.VirtualVariable;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizer;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;
import org.processmining.plugins.balancedconformance.mapping.Variable;
import org.processmining.plugins.balancedconformance.mapping.VariableAwareWriteOperationCost;
import org.processmining.plugins.balancedconformance.result.DataAlignedTrace;

/**
 * {@link DataAlignmentAdapter} that checks if the data alignment could have
 * changed and returns the old alignment in case it is still valid. Otherwise
 * {@link PartialDataAlignedTrace#NO_RESULT} is returned. In case of final
 * variables that are missing or causing a guard expression to be violated
 * {@link PartialDataAlignedTrace#IMPOSSIBLE} is returned.
 * 
 * @author F. Mannhardt
 * 
 */
public final class DataAlignmentAdapterSimpleCheckImpl extends AbstractDataAlignmentAdapter {

	private final static class SimpleAssignmentVisitor extends
			org.processmining.datapetrinets.expression.Evaluator.Visitor {

		private final String varName;

		public SimpleAssignmentVisitor(String varName) {
			this.varName = varName;
		}

		public Object visit(ExprRoot node, Object data) throws ExpressionVisitorException {
			Node child = node.jjtGetChild(0);
			if (child instanceof ExprEqual) {
				return child.jjtAccept(this, data);
			} else {
				return null;
			}
		}

		public Object visit(ExprEqual node, Object data) throws ExpressionVisitorException {
			Node lhsNode = node.jjtGetChild(0);
			Node rhsNode = node.jjtGetChild(1);

			if (lhsNode instanceof ExprVariable && varName.equals(((ExprVariable) lhsNode).jjtGetValue())) {
				Object rhs = rhsNode.jjtAccept(this, data);
				return rhs;
			} else if (rhsNode instanceof ExprVariable && varName.equals(((ExprVariable) rhsNode).jjtGetValue())) {
				Object lhs = lhsNode.jjtAccept(this, data);
				return lhs;
			} else {
				return null;
			}
		}

	}

	private final AbstractPDelegate<? extends Tail> delegate;
	private final DataStateStore dataStateStore;
	private final TraceMapping traceMapping;
	private final double aStarDelta;

	public DataAlignmentAdapterSimpleCheckImpl(LogMapping mapping, TraceMapping traceMapping,
			DataStateFactory dataStateFactory, double aStarDelta, AbstractPDelegate<? extends Tail> delegate,
			DataStateStore dataStateStore) {
		super(mapping, dataStateFactory);
		this.traceMapping = traceMapping;
		this.aStarDelta = aStarDelta;
		this.delegate = delegate;
		this.dataStateStore = dataStateStore;
	}

	public DataAlignedTrace retrieveFullDataAlignment(TraceMapping traceMapping,
			ControlFlowAlignmentResult controlFlowResult) throws DataAlignmentException {
		throw new UnsupportedOperationException(
				"This DataAlignment adapter does not support a complete data alignment!");
	}

	public PartialDataAlignedTrace retrievePartialDataAlignment(TraceMapping traceMapping,
			PartialResultFactory partialResultFactory, DRecord record) throws DataAlignmentException {

		if (hasModelMove(record)) {

			Transition transition = getTransition(record);
			GuardExpression guardExpression = logMapping.getGuardExpression(transition);
			Set<String> variablesToWrite = logMapping.getVariablesToWrite().get(transition);
			boolean hasWriteOps = hasWriteOps(transition);

			try {
				if (hasWriteOps && guardExpression == null) {
					// Just new write operations
					return getAlignmentForWriteOps(transition, record, variablesToWrite);
				} else if (hasWriteOps && guardExpression != null) {
					// New guards and new write operations
					return getDataCostForNewWriteOpsAndGuard(transition, record, partialResultFactory, variablesToWrite);

				} else if (guardExpression != null) {
					// Just a new guard				
					return getDataCostForNewGuardOnly(transition, record);

				} else {
					// No write ops, no guards -> nothing to do
					// Same Cost and Same Data Alignment
					return createSameResult(record);
				}
			} catch (IllegalArgumentException e) {
				// Some problem with the guard
				throw new DataAlignmentException("Could not evaluate guard expression '" + guardExpression.toString()
						+ "' for transition '" + transition + "'! Reason: " + e.getMessage(), e);
			}

		} else {
			// For move in Log, there won't be any data cost 
			// Same Cost and Same Data Alignment
			return createSameResult(record);
		}
	}

	private static boolean hasLogMove(DRecord record) {
		return record.getMovedEvent() != AStarThread.NOMOVE;
	}

	private static boolean hasModelMove(DRecord record) {
		return record.getModelMove() != AStarThread.NOMOVE;
	}

	private PartialDataAlignedTrace createSameResult(DRecord record) {
		boolean didPreviousWrite = hasWriteOps(getTransition(record.getPredecessor()));
		if (didPreviousWrite) {
			// We need to move prime variable
			DataState clonedPreviousVariables = getClonedPreviousVariables(record);
			return new PartialDataAlignedTrace(getPreviousDataCost(record), clonedPreviousVariables);
		} else {
			// There won't be any prime variables that need moving
			return new PartialDataAlignedTrace(getPreviousDataCost(record), record.getPredecessor().getDataState());
		}
	}

	private PartialDataAlignedTrace getDataCostForNewGuardOnly(Transition transition, DRecord record) {
		DataState previousVars = getPreviousVariablesWithMovedPrimes(record);
		GuardExpression guardExpression = logMapping.getGuardExpression(transition);
		if (isGuardTrue(guardExpression, previousVars)) {
			// Same data state
			return createSameResult(record);
		} else {
			if (logMapping.finalVariablesInGuard(transition)) {
				return PartialDataAlignedTrace.IMPOSSIBLE;
			} else {
				return PartialDataAlignedTrace.NO_RESULT;
			}
		}
	}

	private PartialDataAlignedTrace getDataCostForNewWriteOpsAndGuard(Transition transition, DRecord record,
			PartialResultFactory partialResultFactory, Set<String> variablesToWrite) throws DataAlignmentException {

		PartialDataAlignedTrace alignmentForWriteOps = getAlignmentForWriteOps(transition, record, variablesToWrite);

		if (alignmentForWriteOps == PartialDataAlignedTrace.IMPOSSIBLE) {
			return alignmentForWriteOps; // Missing final variables
		} else {

			GuardExpression guardExpression = logMapping.getGuardExpression(transition);
			DataState variableAssignment = alignmentForWriteOps.getVariableAssignment();
			if (variableAssignment == null) {
				// Same result as before
				variableAssignment = traceMapping.getDataStateStore().get(alignmentForWriteOps.getDataStateIndex());
			}

			if (isGuardTrue(guardExpression, variableAssignment)) {
				// Guard is fine, but there may be have been missing write ops
				return alignmentForWriteOps;
			} else {
				// Guard is violated
				if (logMapping.finalVariablesInGuard(transition)) {
					return PartialDataAlignedTrace.IMPOSSIBLE;
				} else {
					// We need to run an ILP solver to check if guard can be satisfied
					return PartialDataAlignedTrace.NO_RESULT;
				}
			}

		}

	}

	private PartialDataAlignedTrace getAlignmentForWriteOps(Transition transition, DRecord record,
			Set<String> variablesToWrite) {

		DataState clonedDataState = null; // lazy init

		int dataCost = 0;

		for (String varName : variablesToWrite) {

			XEvent currentEvent = null;
			if (hasLogMove(record) && hasModelMove(record)) { // Sync move
				currentEvent = calcLastEvent(record);
			}
			Object value = traceMapping.getVariableAssignmentFor(varName, record, currentEvent, transition);

			if (value == null) {

				Variable variable = logMapping.getVariables().get(varName);
				String activityLabel = logMapping.getTransitionToName().get(transition);
				VariableAwareWriteOperationCost variableCost = logMapping.getVariableCost();

				if (variableCost.isFinal(activityLabel, variable)) {
					return PartialDataAlignedTrace.IMPOSSIBLE;
				} else {

					clonedDataState = lazyClone(record, clonedDataState);

					// Missing write operation 
					dataCost = checkedAdd(dataCost,
							convertDataCost(variableCost.notWritingCost(activityLabel, variable), aStarDelta));

					Object defaultValue = determineDefaultValue(transition, clonedDataState, variablesToWrite, varName,
							variable);
					DataStateHelper.storeValue(clonedDataState, DataStateHelper.addPrime(varName), defaultValue,
							variable.getType(), logMapping.getStringDiscretizer());
				}
			} else {

				if (value != VirtualVariable.UNCHANGED) { // Special value used only by final virtual variables!
					Variable variable = logMapping.getVariables().get(varName);
					clonedDataState = lazyClone(record, clonedDataState);
					DataStateHelper.storeValue(clonedDataState, DataStateHelper.addPrime(varName), value,
							variable.getType(), logMapping.getStringDiscretizer());
				} else {
					// Final variable is unchanged, so we don't process a new writing for efficiency reasons
				}

			}
		}

		if (clonedDataState == null) {
			// Nothing changed
			return createSameResult(record);
		} else {
			return new PartialDataAlignedTrace(getPreviousDataCost(record) + dataCost, clonedDataState);
		}
	}

	private XEvent calcLastEvent(DRecord record) {
		//		int executedEvent = record.getExecutedEvent();
		assert record.getExecutedEvents() > 0 : "No event executed yet";
		assert record.getMovedEvent() != AStarThread.NOMOVE : "Unknown event for record";

		return traceMapping.getMappedTrace().get(record.getMovedEvent());
	}

	private int checkedAdd(int dataCost, int newCost) {
		long sum = (long) dataCost + newCost;
		if (sum > Integer.MAX_VALUE) {
			return Integer.MAX_VALUE;
		} else {
			return (int) sum;
		}
	}

	private DataState lazyClone(DRecord record, DataState dataState) {
		if (dataState == null) {
			dataState = getClonedPreviousVariables(record);
		}
		return dataState;
	}

	private Object determineDefaultValue(Transition transition, DataState newAssignment, Set<String> varToWrite,
			String varName, Variable variable) {
		Object defaultValue = null;
		if (varToWrite.size() == 1 && logMapping.getGuardExpression(transition) != null) {
			// Special case, the guard might be actually just a simple assignment to a single variable
			defaultValue = getSimpleAssignmentValue(logMapping.getGuardExpression(transition), varName, newAssignment,
					logMapping.getStringDiscretizer(), logMapping.getVariables());
		}
		if (defaultValue == null) {
			// Either more than one variables are to be assigned or there is no guard and we are free to choose any value within the bounds
			defaultValue = getDefaultValue(variable);
		}
		return defaultValue;
	}

	private Object getSimpleAssignmentValue(GuardExpression guard, String varName, final DataState newAssignment,
			final StringDiscretizer stringDiscretizer, final Map<String, Variable> variables) {
		try {
			Object defaultValue = guard.visit(new SimpleAssignmentVisitor(varName), new Evaluator.Provider() {

				public VariableProvider getVariableProvider() {
					return new DataStateVariableProvider(newAssignment, stringDiscretizer, LogMapping
							.convertToVariableTypes(variables));
				}

				public FunctionProvider getFunctionProvider() {
					return FunctionProvider.EMPTY_PROVIDER;
				}
			});
			if (defaultValue != null) {
				return defaultValue;
			} else {
				return null;
			}
		} catch (ExpressionVisitorException e) {
			return null;
		}
	}

	private Object getDefaultValue(Variable variable) {
		Object defaultValue = null;
		if (variable.getType() != String.class) {
			defaultValue = variable.getDefaultValue();
		}
		if (defaultValue == null) {
			defaultValue = DataStateHelper.getDefaultValue(variable.getType());
		}
		return defaultValue;
	}

	private DataState getClonedPreviousVariables(DRecord record) {
		DataState prev = getPreviousVariables(record);
		if (prev != DataStateStore.EMPTY_IMMUTABLE_STATE) {
			prev = prev.clone(true);
		} else {
			prev = newDataState();
		}
		return prev;
	}

	private DataState getPreviousVariablesWithMovedPrimes(DRecord record) {
		DataState previousVars;
		if (hasWriteOps(getTransition(record.getPredecessor()))) {
			previousVars = getClonedPreviousVariables(record);
		} else {
			previousVars = getPreviousVariables(record);
		}
		return previousVars;
	}

	private DataState getPreviousVariables(DRecord newRec) {
		if (newRec.getPredecessor() != null) {
			DataState dataState = newRec.getPredecessor().getDataState(dataStateStore);
			if (dataState != null) {
				return dataState;
			}
		}
		return DataStateStore.EMPTY_IMMUTABLE_STATE;
	}

	private static int getPreviousDataCost(DRecord record) {
		DRecord predecessor = record.getPredecessor();
		if (predecessor != null) {
			return predecessor.getDataCost();
		} else {
			return 0;
		}
	}

	private Transition getTransition(DRecord rec) {
		return getAbstractPDelegate().getTransition((short) rec.getModelMove());
	}

	private AbstractPDelegate<? extends Tail> getAbstractPDelegate() {
		return this.delegate;
	}

	private boolean isGuardTrue(GuardExpression guardExpression, DataState dataState) {
		if (dataState == null) {
			// We don't support checking for non-existence of a variable here
			return false;
		}
		return ExpressionEvaluator.failSafeIsTrue(guardExpression, dataState,
				LogMapping.convertToVariableTypes(logMapping.getVariables()), logMapping.getStringDiscretizer());
	}

	private boolean hasWriteOps(Transition transition) {
		return logMapping.hasWriteOps(transition);
	}

	public void close() {
	}

}
