/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.dataflow.datastate;

import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import it.unimi.dsi.fastutil.ints.Int2ObjectMap;
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;

/**
 * {@link DataStateStore} implementation using synchronized FastUtil Maps.
 * 
 * @author F. Mannhardt
 * 
 */
public final class DataStateStoreLockImpl implements DataStateStore {

	private int stateCounter = START_INDEX;

	private final Object2IntMap<DataState> dataStateStore;
	private final Int2ObjectMap<DataState> reverseDataStateStore;
	private final ReentrantReadWriteLock lock = new ReentrantReadWriteLock();
	private final Lock readLock = lock.readLock();
	private final Lock writeLock = lock.writeLock();

	public DataStateStoreLockImpl(int initialCapacity) {
		super();
		dataStateStore = new Object2IntOpenHashMap<>(initialCapacity, 0.5f);
		dataStateStore.defaultReturnValue(NO_ENTRY);
		reverseDataStateStore = new Int2ObjectOpenHashMap<>(initialCapacity, 0.5f);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.balancedconformance.dataflow.DataStateStore
	 * #get(long)
	 */
	public DataState get(int index) {
		if (index == EMPTY_STATE_INDEX) {
			return EMPTY_IMMUTABLE_STATE;
		}
		readLock.lock();
		try {
			return reverseDataStateStore.get(index);
		} finally {
			readLock.unlock();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.balancedconformance.dataflow.DataStateStore
	 * #putIfAbsent
	 * (org.processmining.plugins.balancedconformance.dataflow.DataState)
	 */
	public int putIfAbsent(DataState state) {
		if (state == EMPTY_IMMUTABLE_STATE) {
			return EMPTY_STATE_INDEX;
		}
		state = state.makeStateImmutable();
		readLock.lock();
		try {
			int idx = dataStateStore.getInt(state);
			if (idx != NO_ENTRY) {
				return idx;
			}
		} finally {
			readLock.unlock();
		}
		writeLock.lock();
		try {
			int idx = dataStateStore.getInt(state);
			if (idx != NO_ENTRY) {
				return idx;
			} else {
				int stateId = stateCounter++;
				dataStateStore.put(state, stateId);
				reverseDataStateStore.put(stateId, state);
				return stateId;
			}
		} finally {
			writeLock.unlock();
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.balancedconformance.dataflow.DataStateStore
	 * #size()
	 */
	public int size() {
		readLock.lock();
		try {
			return dataStateStore.size();
		} finally {
			readLock.unlock();
		}
	}

}
