package org.processmining.plugins.balancedconformance.export;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.log.utils.XUtils;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.Alignment.AlignmentStep;
import org.processmining.plugins.DataConformance.ResultReplay;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.visualization.DataAwareStepTypes;
import org.processmining.plugins.balancedconformance.result.AlignmentCollection;
import org.processmining.xesalignmentextension.XAlignmentExtension;
import org.processmining.xesalignmentextension.XAlignmentExtension.MoveType;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignmentMove;
import org.processmining.xesalignmentextension.XDataAlignmentExtension;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.DataMoveType;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.IncorrectXAttribute;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.XDataAlignmentExtensionException;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.XDataAlignmentMove;

import com.google.common.collect.ImmutableList;
import com.google.common.primitives.Doubles;

/**
 * Converter between DataAwareReplayer's {@link Alignment} or
 * {@link ResultReplay} / {@link AlignmentCollection} and the {@link XLog} based
 * representation with {@link XAlignmentExtension} and
 * {@link XDataAlignmentExtension}.
 * 
 * This class is NOT thread-safe!
 * 
 * @author F. Mannhardt
 * 
 */
public final class XAlignmentConverter {

	/**
	 * Lazy view of the {@link AlignmentStep} as an {@link XDataAlignmentMove}.
	 * Only {@link #getEvent()} involves expensive copying.
	 * 
	 * @author F. Mannhardt
	 * 
	 */
	private static final class XDataAlignmentMoveView implements XDataAlignmentExtension.XDataAlignmentMove {

		private final AlignmentStep step;
		private XAlignmentConverter converter;

		private XDataAlignmentMoveView(final AlignmentStep step, final XAlignmentConverter converter) {
			this.step = step;
			this.converter = converter;
		}

		@Override
		public boolean isObservable() {
			return step.getType() != DataAwareStepTypes.MINVI;
		}

		@Override
		public MoveType getType() {
			return convertType(step.getType());
		}

		@Override
		public String getModelMove() {
			return step.getProcessView().getActivity();
		}

		@Override
		public String getLogMove() {
			return step.getLogView().getActivity();
		}

		@Override
		public String getEventClassId() {
			return converter.convertEventClassId(step.getLogView());
		}

		@Override
		public String getActivityId() {
			return converter.convertTransitionId(step.getProcessView());
		}

		@Override
		public DataMoveType getDataMoveType() {
			return convertDataMoveType(step);
		}

		@Override
		public List<XAttribute> getCorrectAttributes() throws XDataAlignmentExtensionException {
			return converter.convertCorrectWritings(step);
		}

		@Override
		public List<IncorrectXAttribute> getIncorrectAttributes() throws XDataAlignmentExtensionException {
			return converter.convertIncorrectWritings(step);
		}

		@Override
		public List<XAttribute> getMissingAttributes() throws XDataAlignmentExtensionException {
			return converter.convertMissingWritings(step);
		}

		@Override
		public List<XAttribute> getUnmappedAttributes() {
			return converter.convertUnmappedAttributes(step);
		}

		/**
		 * CAREFUL: This is results in a full copy of the alignment step as a
		 * {@link XEvent}.
		 * 
		 * @see org.processmining.plugins.balancedconformance.export.XAlignmentExtension.XAlignmentMove#getEvent()
		 */
		@Override
		public XEvent getEvent() {
			// Expensive copy
			return converter.copyAsXEvent(step);
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = (prime * result) + ((step == null) ? 0 : step.hashCode());
			return result;
		}

		@Override
		public boolean equals(final Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			final XDataAlignmentMoveView other = (XDataAlignmentMoveView) obj;
			if (step == null) {
				if (other.step != null) {
					return false;
				}
			} else if (!step.equals(other.step)) {
				return false;
			}
			return true;
		}

	}

	/**
	 * Lazy view of the {@link Alignment} as an {@link XAlignment}. Only
	 * {@link #getTrace()} involves expensive copying.
	 * 
	 * @author F. Mannhardt
	 * 
	 */
	private static final class XAlignmentView extends AbstractList<XAlignmentMove> implements XAlignment {

		private final Alignment a;
		private final XTrace originalTrace;

		private final XAlignmentConverter converter;

		private XAlignmentView(final Alignment alignment, final XTrace originalTrace, XAlignmentConverter converter) {
			this.a = alignment;
			this.originalTrace = originalTrace;
			this.converter = converter;
		}

		@Override
		public int size() {
			return a.getStepTypes().size();
		}

		@Override
		public double getFitness() {
			return a.getFitness();
		}

		/**
		 * CAREFUL: This is results in a full copy of the alignment as a
		 * {@link XTrace}.
		 * 
		 * @see org.processmining.plugins.balancedconformance.export.XAlignmentExtension.XAlignment#getTrace()
		 */
		@Override
		public XTrace getTrace() {
			// This is expensive, and not a 'view' anymore
			return converter.copyAsXTrace(a, originalTrace);
		}

		@Override
		public String getName() {
			return a.getTraceName();
		}

		@Override
		public XDataAlignmentMoveView get(final int index) {
			final ExecutionStep logStep = a.getLogTrace().get(index);
			final ExecutionStep processStep = a.getProcessTrace().get(index);
			final DataAwareStepTypes type = a.getStepTypes().get(index);
			return new XDataAlignmentMoveView(new Alignment.AlignmentStepImpl(type, processStep, logStep), converter);
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = (prime * result) + ((getName() == null) ? 0 : getName().hashCode());
			result = (prime * result) + (Doubles.hashCode(getFitness()));
			return result;
		}

		@Override
		public boolean equals(final Object obj) {
			if (this == obj) {
				return true;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			final XAlignmentView other = (XAlignmentView) obj;
			if (getName() == null) {
				if (other.getName() != null) {
					return false;
				}
			} else if (!getName().equals(other.getName())) {
				return false;
			}
			if (Double.compare(getFitness(), other.getFitness()) != 0) {
				return false;
			}
			return true;
		}

	}

	private XFactory factory = XFactoryRegistry.instance().currentDefault();
	private XEventClassifier classifier = null;
	private Map<String, String> variableMapping = null;

	public XAlignmentConverter() {
		super();
	}

	/**
	 * Returns a view, mostly lazily computed, on {@link Alignment} as a
	 * {@link XAlignment}.
	 * 
	 * @param alignment
	 * @return
	 */
	public XAlignmentView viewAsXAlignment(final Alignment alignment) {
		return viewAsXAlignment(alignment, null);
	}

	public XAlignmentView viewAsXAlignment(final Alignment alignment, final XTrace originalTrace) {
		return new XAlignmentView(alignment, originalTrace, this);

	}

	public XAlignment copyAsXAlignment(final Alignment alignment) {
		return XAlignmentExtension.instance().extendTrace(copyAsXTrace(alignment));
	}

	public XAlignment copyAsXAlignment(final Alignment alignment, final XTrace originalTrace) {
		return XAlignmentExtension.instance().extendTrace(copyAsXTrace(alignment, originalTrace));
	}

	public XTrace copyAsXTrace(final Alignment alignment) {
		return copyAsXTrace(alignment, null);
	}

	public XTrace copyAsXTrace(final Alignment alignment, final XTrace originalTrace) {

		final XAlignmentExtension alignmentExtension = XAlignmentExtension.instance();

		XTrace newTrace;
		if (originalTrace != null) {
			newTrace = factory.createTrace();
			newTrace.setAttributes((XAttributeMap) originalTrace.getAttributes().clone());
		} else {
			newTrace = factory.createTrace();
			final String traceName = alignment.getTraceName();
			XConceptExtension.instance().assignName(newTrace, traceName);
		}

		alignmentExtension.assignFitness(newTrace, alignment.getFitness());

		final Iterator<AlignmentStep> iter = alignment.alignmentStepIterator();
		while (iter.hasNext()) {
			newTrace.add(copyAsXEvent(iter.next()));
		}
		return newTrace;
	}

	public XEvent copyAsXEvent(final Alignment.AlignmentStep step) {

		XEvent event = factory.createEvent();

		assignTypeOfMove(step, event);

		if (step.getType() != DataAwareStepTypes.L) { // Model has moved
			assignModelMoveAttributes(step, event);
		} else {
			assignLogMoveAttributes(step, event);
		}

		return event;
	}

	private void assignModelMoveAttributes(final Alignment.AlignmentStep step, XEvent event) {
		int variableCount = step.getProcessView().size();
		List<XAttribute> correctAttributes = new ArrayList<>(variableCount);
		List<XAttribute> inCorrectAttributes = new ArrayList<>(variableCount);
		List<XAttribute> missingAttributes = new ArrayList<>(variableCount);

		bulkConvertAttributes(step, correctAttributes, inCorrectAttributes, missingAttributes);

		// Without concept:name!!
		List<XAttribute> unmappedAttributes = convertUnmappedAttributes(step);

		XUtils.putAttributes(event, correctAttributes);
		XUtils.putAttributes(event, missingAttributes);
		XUtils.putAttributes(event, unmappedAttributes);
		XUtils.putAttributes(event, inCorrectAttributes);

		// Override concept:name from the model so that the log is 'aligned', original event class is stored
		XConceptExtension.instance().assignName(event, step.getProcessView().getActivity());
	}

	private void assignLogMoveAttributes(final Alignment.AlignmentStep step, XEvent event) {
		// Log move just copy original attributes if available
		if (step.getLogView().getActivityObject() instanceof XEvent) {
			XAttributeMap originalAttributes = ((XEvent) step.getLogView().getActivityObject())
					.getAttributes();
			for (XAttribute attr: originalAttributes.values()) {
				event.getAttributes().put(attr.getKey(), (XAttribute) attr.clone());
			}
		}
	}

	private void assignTypeOfMove(final Alignment.AlignmentStep step, final XEvent event) {

		final XAlignmentExtension alignmentExtension = XAlignmentExtension.instance();
		final XDataAlignmentExtension dataAlignmentExtension = XDataAlignmentExtension.instance();

		if ((step.getType() != DataAwareStepTypes.MINVI) && (step.getType() != DataAwareStepTypes.MREAL)) {
			// Log Move | Synchronous
			if (step.getType() == DataAwareStepTypes.L) {
				alignmentExtension.assignLogMove(event, step.getLogView().getActivity());
				if (classifier != null) {
					alignmentExtension.assignEventClassId(event, convertEventClassId(step.getLogView()));
				}
				alignmentExtension.assignIsObservable(event, true);
			} else {
				alignmentExtension.assignLogMove(event, step.getLogView().getActivity());
				if (classifier != null) {
					alignmentExtension.assignEventClassId(event, convertEventClassId(step.getLogView()));
				}
				alignmentExtension.assignModelMove(event, step.getProcessView().getActivity());
				alignmentExtension.assignActivityId(event, convertTransitionId(step.getProcessView()));
				alignmentExtension.assignIsObservable(event, true);
			}
		} else {
			// Model Move
			alignmentExtension.assignModelMove(event, step.getProcessView().getActivity());
			alignmentExtension.assignActivityId(event, convertTransitionId(step.getProcessView()));
			if (step.getType() == DataAwareStepTypes.MINVI) {
				alignmentExtension.assignIsObservable(event, false);
			} else {
				alignmentExtension.assignIsObservable(event, true);
			}
		}

		alignmentExtension.assignMoveType(event, convertType(step.getType()));
		dataAlignmentExtension.assignDataMoveType(event, convertDataMoveType(step));
	}

	private void bulkConvertAttributes(final Alignment.AlignmentStep step, final List<XAttribute> correctAttributes,
			final List<XAttribute> inCorrectAttributes, final List<XAttribute> missingAttributes) {
		for (final Iterator<Entry<String, Object>> variableIter = step.getProcessView().entrySet()
				.iterator(); variableIter.hasNext();) {

			final Entry<String, Object> variableWriting = variableIter.next();

			final Object modelValue = variableWriting.getValue();
			final Object logValue = step.getLogView() != null ? step.getLogView().get(variableWriting.getKey()) : null;

			if (logValue == null) {

				final XAttribute modelAttribute = createModelAttribute(variableWriting, variableWriting.getValue(),
						variableMapping);
				XDataAlignmentExtension.instance().assignDataMoveType(modelAttribute, DataMoveType.MISSING);
				missingAttributes.add(modelAttribute);

			} else if (modelValue.equals(logValue)) {

				final XAttribute modelAttribute = createModelAttribute(variableWriting, logValue, variableMapping);
				XDataAlignmentExtension.instance().assignDataMoveType(modelAttribute, DataMoveType.CORRECT);
				correctAttributes.add(modelAttribute);

			} else if (!modelValue.equals(logValue)) {

				final XAttribute modelAttribute = createModelAttribute(variableWriting, logValue, variableMapping);
				XDataAlignmentExtension.instance().assignDataMoveType(modelAttribute, DataMoveType.INCORRECT);
				inCorrectAttributes.add(modelAttribute);

			} else {
				throw new IllegalStateException();
			}

		}
	}

	private List<XAttribute> convertCorrectWritings(final AlignmentStep step) {

		final List<XAttribute> correctAttributes = new ArrayList<>();
		for (final Iterator<Entry<String, Object>> variableIter = step.getProcessView().entrySet()
				.iterator(); variableIter.hasNext();) {

			final Entry<String, Object> variableWriting = variableIter.next();

			final Object modelValue = variableWriting.getValue();
			final Object logValue = step.getLogView() != null ? step.getLogView().get(variableWriting.getKey()) : null;

			if (modelValue.equals(logValue)) {

				final XAttribute modelAttribute = createModelAttribute(variableWriting, logValue, variableMapping);
				XDataAlignmentExtension.instance().assignDataMoveType(modelAttribute, DataMoveType.CORRECT);
				correctAttributes.add(modelAttribute);

			}
		}
		return correctAttributes;

	}

	private List<IncorrectXAttribute> convertIncorrectWritings(final AlignmentStep step) {

		final List<IncorrectXAttribute> inCorrectAttributes = new ArrayList<>();

		for (final Iterator<Entry<String, Object>> variableIter = step.getProcessView().entrySet()
				.iterator(); variableIter.hasNext();) {

			final Entry<String, Object> variableWriting = variableIter.next();

			final Object modelValue = variableWriting.getValue();
			final Object logValue = step.getLogView() != null ? step.getLogView().get(variableWriting.getKey()) : null;

			if (logValue != null) {
				// Incorrect writing

				if (!modelValue.equals(logValue)) {

					final XAttribute modelAttribute = createModelAttribute(variableWriting, logValue, variableMapping);
					final XAttribute logAttribute = XDataAlignmentExtension.instance().extractLogValue(modelAttribute);
					XDataAlignmentExtension.instance().assignDataMoveType(modelAttribute, DataMoveType.INCORRECT);

					inCorrectAttributes.add(new IncorrectXAttribute() {

						@Override
						public XAttribute getModelAttribute() {
							return modelAttribute;
						}

						@Override
						public XAttribute getLogAttribute() {
							return logAttribute;
						}
					});
				}
			}
		}
		return inCorrectAttributes;
	}

	private List<XAttribute> convertMissingWritings(final AlignmentStep step) {

		final List<XAttribute> missingAttributes = new ArrayList<>();

		for (final Iterator<Entry<String, Object>> variableIter = step.getProcessView().entrySet()
				.iterator(); variableIter.hasNext();) {

			final Entry<String, Object> variableWriting = variableIter.next();

			final Object logValue = step.getLogView() != null ? step.getLogView().get(variableWriting.getKey()) : null;

			if (logValue == null) {

				final XAttribute modelAttribute = createModelAttribute(variableWriting, variableWriting.getValue(),
						variableMapping);
				XDataAlignmentExtension.instance().assignDataMoveType(modelAttribute, DataMoveType.MISSING);
				missingAttributes.add(modelAttribute);

			}
		}
		return missingAttributes;

	}

	public List<XAttribute> convertUnmappedAttributes(final AlignmentStep step) {

		List<XAttribute> unmappedAttributes = new ArrayList<>();

		if (step.getLogView().getActivityObject() instanceof XEvent) {

			XAttributeMap attributes = ((XEvent) step.getLogView().getActivityObject()).getAttributes();
			Set<String> mappedAttributes = new HashSet<>();
			mappedAttributes.add("concept:name");

			if (step.getProcessView() != null && variableMapping != null) {
				for (String varName : step.getProcessView().keySet()) {
					mappedAttributes.add(variableMapping.get(varName));
				}
			}

			for (XAttribute a : attributes.values()) {
				if (!mappedAttributes.contains(a.getKey())) {
					unmappedAttributes.add(a); //TODO this does not clone ?!
				}
			}
			return unmappedAttributes;

		} else {
			return ImmutableList.of();
		}
	}

	private XAttribute createModelAttribute(final Entry<String, Object> variableWriting, final Object logValue,
			final Map<String, String> variableMapping) {

		final String modelAttributeName = variableWriting.getKey();
		String logAttributeName = variableMapping != null ? variableMapping.get(variableWriting.getKey())
				: variableWriting.getKey();
		if (logAttributeName == null) {
			// Mapping is missing artificial variables
			logAttributeName = variableWriting.getKey();
		}

		final XAttribute modelAttribute = createAttribute(modelAttributeName, variableWriting.getValue(),
				XDataAlignmentExtension.instance());
		if (logAttributeName != null) {
			XDataAlignmentExtension.instance().assignLogAttribute(modelAttribute, logAttributeName, logValue);
		}
		XDataAlignmentExtension.instance().assignIsModelAttribute(modelAttribute);
		return modelAttribute;
	}

	/**
	 * Creates an appropriate {@link XAttribute}, deciding by the type of the
	 * parameter atttributeValue.
	 * 
	 * @param attributeName
	 * @param attributeValue
	 * @param extension
	 * @return a {@link XAttribute} with correct type
	 */
	private XAttribute createAttribute(String attributeName, Object attributeValue, XExtension extension) {
		if (attributeValue instanceof Double || attributeValue instanceof Float) {
			return factory.createAttributeContinuous(attributeName, ((Number) attributeValue).doubleValue(), extension);
		} else if (attributeValue instanceof Integer || attributeValue instanceof Long) {
			return factory.createAttributeDiscrete(attributeName, ((Number) attributeValue).longValue(), extension);
		} else if (attributeValue instanceof Date) {
			return factory.createAttributeTimestamp(attributeName, ((Date) attributeValue), extension);
		} else if (attributeValue instanceof Boolean) {
			return factory.createAttributeBoolean(attributeName, ((Boolean) attributeValue), extension);
		} else {
			return factory.createAttributeLiteral(attributeName, attributeValue.toString(), extension);
		}
	}

	private String convertTransitionId(final ExecutionStep processView) {
		if (processView.getActivityObject() instanceof Transition) {
			Transition transition = (Transition) processView.getActivityObject();
			return tryGetLocalIdFromCache(transition);
		} else {
			return null;
		}
	}

	private final Map<Transition, String> localIdCache = new HashMap<>();

	private String tryGetLocalIdFromCache(Transition transition) {
		String localId = localIdCache.get(transition);
		if (localId == null) {
			localId = transition.getLocalID().localNodeID().toString();
			localIdCache.put(transition, localId);
		}
		return localId;
	}

	private String convertEventClassId(final ExecutionStep logView) {
		if (logView.getActivityObject() instanceof XEvent) {
			return classifier.getClassIdentity((XEvent) logView.getActivityObject());
		} else {
			return null;
		}
	}

	private static DataMoveType convertDataMoveType(final AlignmentStep step) {
		if (step.getType() == DataAwareStepTypes.MREAL && !step.getProcessView().isEmpty()) {
			return DataMoveType.MISSING;
		}
		if (step.getType() == DataAwareStepTypes.MINVI) {
			if (step.getLogView() != ExecutionStep.bottomStep) {
				// Data has moved
				boolean correctData = step.getProcessView().equalsAssignmentsOnly(step.getLogView());
				return correctData ? DataMoveType.CORRECT : DataMoveType.INCORRECT;
			}
		}
		if ((step.getType() == DataAwareStepTypes.LMNOGOOD)) {
			if (step.hasMissingVariableAssignment()) {
				return DataMoveType.MISSING;
			} else {
				return DataMoveType.INCORRECT;
			}
		}
		return DataMoveType.CORRECT;
	}

	private static XAlignmentExtension.MoveType convertType(final DataAwareStepTypes type) {
		switch (type) {
			case L :
				return MoveType.LOG;
			case LMGOOD :
				return MoveType.SYNCHRONOUS;
			case LMNOGOOD :
				return MoveType.SYNCHRONOUS;
			case MINVI :
			case MREAL :
				return MoveType.MODEL;
			default :
				throw new IllegalArgumentException("Invalid move type " + type);

		}
	}

	public XFactory getFactory() {
		return factory;
	}

	public XAlignmentConverter setFactory(XFactory factory) {
		this.factory = factory;
		return this;
	}

	public XEventClassifier getClassifier() {
		return classifier;
	}

	public XAlignmentConverter setClassifier(XEventClassifier classifier) {
		this.classifier = classifier;
		return this;
	}

	public Map<String, String> getVariableMapping() {
		return variableMapping;
	}

	public XAlignmentConverter setVariableMapping(Map<String, String> variableMapping) {
		this.variableMapping = variableMapping;
		return this;
	}

}
