package org.processmining.plugins.balancedconformance.functions.impl;

import java.util.Date;
import java.util.Set;

import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XEvent;
import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;
import org.processmining.datapetrinets.expression.syntax.ExprFunction;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.DataConformance.framework.WriteOperationCost;
import org.processmining.plugins.balancedconformance.functions.FunctionUnary;
import org.processmining.plugins.balancedconformance.functions.VirtualVariable;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.Variable;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSetMultimap;
import com.google.common.collect.SetMultimap;

final class HourOfDayFunction implements FunctionUnary {

	final static class HourOfDayVariable implements VirtualVariable {

		private final Transition transition;
		private final DateTimeZone timeZone;

		private String name = "hourOfDay";
		private Class<?> type = Long.class;

		public HourOfDayVariable(Transition transition, String timeZone) {
			this.transition = transition;
			if (timeZone == null) {
				this.timeZone = DateTimeZone.getDefault();
			} else {
				this.timeZone = DateTimeZone.forID(timeZone);
			}
		}

		public Object compute(Prefix alignmentPrefix, Transition currentTransition, XEvent currentEvent,
				LogMapping mapping) {
			if (currentEvent != null) {
				Date currentTime = XTimeExtension.instance().extractTimestamp(currentEvent);
				if (currentTime != null) {
					DateTime dateTime = new DateTime(currentTime.getTime(), timeZone);
					int hourOfDay = dateTime.getHourOfDay();
					return Long.valueOf(hourOfDay);
				}
			}
			return null;
		}

		public String getName() {
			return name;
		}

		public Class<?> getType() {
			return type;
		}

		public Set<String> getRelevantAttributes() {
			return ImmutableSet.of(XTimeExtension.KEY_TIMESTAMP);
		}

		public Set<Transition> getRelevantTransitions() {
			return ImmutableSet.of(transition);
		}

		public int getFaultyValueCost() {
			return Integer.MAX_VALUE;
		}

		public int getMissingValueCost() {
			return Integer.MAX_VALUE; // Should never be relevant
		}

		public Double getUpperBound() {
			return null; // Does not matter as this will never be part of an ILP problem
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((getName() == null) ? 0 : getName().hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (!(obj instanceof HourOfDayVariable))
				return false;
			HourOfDayVariable other = (HourOfDayVariable) obj;
			if (getName() == null) {
				if (other.getName() != null)
					return false;
			} else if (!getName().equals(other.getName()))
				return false;
			return true;
		}

		public boolean isFinal() {
			return true;
		}

		public boolean needsAlignedPrefix() {
			return false;
		}

	}

	public boolean isValidFunction(ExprFunction node) {
		return node.jjtGetNumChildren() >= 0;
	}

	public VirtualVariable createPrimaryVariable(ImmutableMap<String, Variable> normalVariables,
			WriteOperationCost writeOperationCost, SetMultimap<String, Transition> nameToTransition,
			Transition transition, String parameter1) {
		return new HourOfDayVariable(transition, parameter1);
	}

	public SetMultimap<Transition, VirtualVariable> createHelperVariables(
			ImmutableMap<String, Variable> normalVariables, WriteOperationCost writeOperationCost,
			SetMultimap<String, Transition> nameToTransition, Transition transition,
			VirtualVariable primaryVariable, String parameter1) {
		return ImmutableSetMultimap.of();
	}

}