package org.processmining.plugins.balancedconformance.functions.impl;

import java.util.Collection;
import java.util.Date;
import java.util.Set;

import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XEvent;
import org.processmining.datapetrinets.expression.syntax.ExprFunction;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.VariableAccess;
import org.processmining.models.graphbased.directed.utils.GraphIterator;
import org.processmining.models.graphbased.directed.utils.GraphIterator.EdgeAcceptor;
import org.processmining.models.graphbased.directed.utils.GraphIterator.NodeAcceptor;
import org.processmining.plugins.DataConformance.framework.WriteOperationCost;
import org.processmining.plugins.balancedconformance.functions.FunctionBinary;
import org.processmining.plugins.balancedconformance.functions.FunctionException;
import org.processmining.plugins.balancedconformance.functions.VirtualVariable;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.Variable;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.SetMultimap;

final class TimeSinceFunction implements FunctionBinary {

	private static final String EARLIEST_TIME = "EarliestTime";
	private static final String LATEST_TIME = "LatestTime";

	final static class VirtualTimeVariableImpl implements VirtualVariable {

		private ImmutableSet<Transition> transitionsBetween;

		public VirtualTimeVariableImpl(final Transition transition, final Transition referenceTransition,
				String timeResolution) {
			setName(transformToIdentifier(referenceTransition.getLabel()));
			Collection<PetrinetNode> predecessors = calcTransitionsInBetween(transition, referenceTransition);
			this.transitionsBetween = ImmutableSet.copyOf(predecessors.toArray(new Transition[0]));
			if (timeResolution != null) {
				resolution = TimeResolution.fromUnit(timeResolution);
			}
			earliestTimeVariable = getName() + EARLIEST_TIME;
			latestTimeVariable = getName() + LATEST_TIME;
		}

		private Collection<PetrinetNode> calcTransitionsInBetween(final Transition transition,
				final Transition referenceTransitions) {
			Collection<PetrinetNode> predecessors = GraphIterator.getDepthFirstPredecessors(transition,
					transition.getGraph(),
					new EdgeAcceptor<PetrinetNode, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>>() {
						public boolean acceptEdge(PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge,
								int depth) {
							return !(edge instanceof VariableAccess || referenceTransitions == edge.getSource());
						}
					}, new NodeAcceptor<PetrinetNode>() {
						public boolean acceptNode(PetrinetNode node, int depth) {
							return (node instanceof Transition) && !(depth == 0 || ((Transition) node).isInvisible());
						}
					});
			predecessors.add(transition);
			predecessors.add(referenceTransitions);
			return predecessors;
		}

		private String transformToIdentifier(String transitionName) {
			return "timeSince_" + escapeVariableName(transitionName);
		}

		private String escapeVariableName(String transitionName) {
			return transitionName.replace(" ", "").replace("-", "").replace("@", "").replace("+", ""); //TODO make a proper implementation!
		}

		private String name;
		private TimeResolution resolution = TimeResolution.SECONDS;
		private Class<?> type = Long.class;
		private String latestTimeVariable;
		private String earliestTimeVariable;

		public String getName() {
			return name;
		}

		public void setName(String name) {
			this.name = name;
		}

		public TimeResolution getResolution() {
			return resolution;
		}

		public Class<?> getType() {
			return type;
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((getName() == null) ? 0 : getName().hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (!(obj instanceof VirtualTimeVariableImpl))
				return false;
			VirtualTimeVariableImpl other = (VirtualTimeVariableImpl) obj;
			if (getName() == null) {
				if (other.getName() != null)
					return false;
			} else if (!getName().equals(other.getName()))
				return false;
			return true;
		}

		public Object compute(Prefix prefix, Transition currentTransition, XEvent currentEvent, LogMapping mapping) {

			assert prefix.isAligned();

			// Both are evaluated beforehand, so those will contain the correct values
			Long earliestTime = (Long) prefix.variableValue(earliestTimeVariable);
			Long latestTime = (Long) prefix.variableValue(latestTimeVariable);

			if (currentTransition != null && currentEvent != null) {
				//synchronous move on exit transition - update latest time as we cannot yet read prime variables here
				XAttribute timeAttribute = currentEvent.getAttributes()
						.get(VirtualVariable.ATTRIBUTE_KEY_RELATIVE_TIME);
				if (timeAttribute != null) {
					Date timestamp = ((XAttributeTimestamp) timeAttribute).getValue();
					latestTime = Long.valueOf(timestamp.getTime()); // we are the latest
				}
			}

			if (earliestTime == null || latestTime == null) {
				// No move with a time stamp occurred between 
				return 0l;
			} else {

				if (latestTime == -1) {
					return 0l; // nothing happened after reference transition, and the source transition does not have a time stamp 
				}

				if (earliestTime == -1) { // no visible event occurred after the reference transition 
					// we have to look at the last time before the reference transition occurred
					//TODO find a way to iterate backwards! 
					//TODO find a way to do this nicely with a third variable
					Iterable<PrefixStep> steps = prefix.getSteps();
					long lastTimeBefore = 0;
					for (PrefixStep step : steps) {
						if (step.getProcessObject() != null && step.getLogObject() != null) { // Sync move
							XEvent event = step.getLogObject();
							Date timestamp = ((XAttributeTimestamp) event.getAttributes()
									.get(VirtualVariable.ATTRIBUTE_KEY_RELATIVE_TIME)).getValue();
							if (timestamp != null) {
								lastTimeBefore = Long.valueOf(timestamp.getTime());
							}
						}
					}
					// we take the last known time an event occurred beforehand as earliest time
					earliestTime = lastTimeBefore;
				}

				assert latestTime >= earliestTime : "Latest time should be equal or larger than earliest time for timeSince constraint!";
				double newTime = (latestTime - earliestTime) / (double) getResolution().getFactor();
				return Long.valueOf((long) Math.ceil(newTime));
			}

		}

		public Set<Transition> getRelevantTransitions() {
			return transitionsBetween;
		}

		public Set<String> getRelevantAttributes() {
			return ImmutableSet.of(); // only the helper variables actually use attributes from the log
		}

		public int getFaultyValueCost() {
			return Integer.MAX_VALUE;
		}

		public int getMissingValueCost() {
			return Integer.MAX_VALUE; // Should never be relevant
		}

		public Double getUpperBound() {
			return null; // Does not matter as this will never be part of an ILP problem
		}

		public boolean isFinal() {
			return true;
		}

		public boolean needsAlignedPrefix() {
			return true;
		}

	}

	final static class EarliestTimestampVariable implements VirtualVariable {

		private final String pattern;
		private final Transition exitTransition;

		public EarliestTimestampVariable(String pattern, Transition exitTransition) {
			this.exitTransition = exitTransition;
			this.pattern = pattern + EARLIEST_TIME;
		}

		public Double getUpperBound() {
			return null;
		}

		public Class<?> getType() {
			return Long.class;
		}

		public Set<Transition> getRelevantTransitions() {
			return ImmutableSet.of(); // only the current transition is relevant
		}

		public Set<String> getRelevantAttributes() {
			return ImmutableSet.of(); // only latest timestamp variable is relevant
		}

		public String getName() {
			return pattern;
		}

		public int getFaultyValueCost() {
			return Integer.MAX_VALUE;
		}

		public int getMissingValueCost() {
			return Integer.MAX_VALUE;
		}

		public Object compute(Prefix prefix, Transition currentTransition, XEvent currentEvent, LogMapping mapping) {

			Long availableValue = (Long) prefix.variableValue(getName());
			if (availableValue == null || availableValue == -1) {

				if (currentTransition != null && currentEvent != null) { //synchronous move
					XAttribute timeAttribute = currentEvent.getAttributes()
							.get(VirtualVariable.ATTRIBUTE_KEY_RELATIVE_TIME);
					if (timeAttribute != null) {
						Date timestamp = ((XAttributeTimestamp) timeAttribute).getValue();
						return Long.valueOf(timestamp.getTime());
					}
				}

				// No time stamp know
				return -1;

			} else {

				if (exitTransition == currentTransition) {
					return -1; // reset for second run
				} else {
					// Some other event beforehand already calculated the value, so we know the earliest time stamp
					return availableValue;
				}
			}
		}

		public boolean isFinal() {
			return true;
		}

		public boolean needsAlignedPrefix() {
			return true;
		}

	}

	final static class LatestTimestampVariable implements VirtualVariable {

		private final String pattern;
		private final Set<Transition> entryTransitions;
		private final Transition entryTransition;

		public LatestTimestampVariable(String pattern, final Set<Transition> referenceTransitions,
				Transition transition) {
			this.entryTransitions = referenceTransitions;
			if (referenceTransitions.size() == 1) {
				this.entryTransition = referenceTransitions.iterator().next();
			} else {
				this.entryTransition = null;
			}
			this.pattern = pattern + LATEST_TIME;
		}

		public Double getUpperBound() {
			return null;
		}

		public Class<?> getType() {
			return Long.class;
		}

		public Set<Transition> getRelevantTransitions() {
			return ImmutableSet.of(); // only the current transition is relevant
		}

		public Set<String> getRelevantAttributes() {
			return ImmutableSet.of(VirtualVariable.ATTRIBUTE_KEY_RELATIVE_TIME); // workaround to signal the grouping to use relative time
		}

		public String getName() {
			return pattern;
		}

		public int getFaultyValueCost() {
			return Integer.MAX_VALUE;
		}

		public int getMissingValueCost() {
			return Integer.MAX_VALUE;
		}

		public Object compute(Prefix prefix, Transition currentTransition, XEvent currentEvent, LogMapping mapping) {

			if (currentTransition != null && currentEvent != null) { //synchronous move
				XAttribute timeAttribute = currentEvent.getAttributes()
						.get(VirtualVariable.ATTRIBUTE_KEY_RELATIVE_TIME);
				if (timeAttribute != null) {
					Date timestamp = ((XAttributeTimestamp) timeAttribute).getValue();
					return Long.valueOf(timestamp.getTime());
				}
			} else if ((entryTransition != null && currentTransition == entryTransition)
					|| entryTransitions.contains(currentTransition)) {
				return -1; // reset for second run
			}

			return UNCHANGED;
		}

		public boolean isFinal() {
			return true;
		}

		public boolean needsAlignedPrefix() {
			return false;
		}

	}

	public boolean isValidFunction(ExprFunction node) {
		return node.jjtGetNumChildren() >= 1;
	}

	public VirtualVariable createPrimaryVariable(ImmutableMap<String, Variable> normalVariables,
			WriteOperationCost writeOperationCost, SetMultimap<String, Transition> nameToTransition,
			Transition transition, String parameter1, String parameter2) throws FunctionException {

		Set<Transition> referenceTransitions = nameToTransition.get(parameter1);

		if (referenceTransitions.size() != 1) {
			throw new FunctionException(String.format(
					"Function 'timeSince' is referencing an unkown or duplicate (ambigous) transition %s", parameter1));
		}

		return new VirtualTimeVariableImpl(transition, referenceTransitions.iterator().next(), parameter2);
	}

	public SetMultimap<Transition, VirtualVariable> createHelperVariables(
			ImmutableMap<String, Variable> normalVariables, WriteOperationCost writeOperationCost,
			SetMultimap<String, Transition> nameToTransition, Transition transition, VirtualVariable primaryVariable,
			String parameter1, String parameter2) {

		Set<Transition> referenceTransitions = nameToTransition.get(parameter1);

		SetMultimap<Transition, VirtualVariable> helperVariables = LinkedHashMultimap.create();
		for (final Transition t : primaryVariable.getRelevantTransitions()) {
			helperVariables.put(t, new EarliestTimestampVariable(primaryVariable.getName(), transition));
			helperVariables.put(t,
					new LatestTimestampVariable(primaryVariable.getName(), referenceTransitions, transition));
		}

		return helperVariables;

	}

}